/*

trawl through firmware ~(0x1da2, 7586) functions

replace all 'ret'
801704a8 03 e0 00 08                             jr      $ra
with
801704a8 08 xx xx xx                             j       profiler


profiler
    caller_addr = $ra-8
    callee_addr = *($ra-8)
    match caller_addr, callee_addr
    increment count

+
count the process changes

*/
//
//requires libfirebird
//
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "libFirebird.h"
#include "hash.h"

#ifdef _WIN32
#else
#define ID_USERTAP 0xbdb02b01
TAP_ID( ID_USERTAP );
TAP_PROGRAM_NAME("profiler");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("v0.01 profiler");
TAP_ETCINFO(__DATE__);
#endif

//----------------------------------------------------------------------------
//
//firmware dependant
//

#define CODE_START  0x80000000          //13.59
#define CODE_END    0x80257444          //13.59

//
//hooking these causes a crash ... 13.59 ...
//
dword exclude[] = {
    0x80000414,                     //lock
    0x80000434,                     //unlock
    0x800026d0,                     //get_meminfo
    0x80002c60,                     //malloc
    0x80178ca0,                     //(TAP_MemAlloc)
    0x80178d20,                     //(TAP_MemFree)
    0x8017afb4,                     //TAP_MemAlloc
};

//for testing
dword include[] = {
//    0x80108384,     //called from TAP_PlayPCM
//    0x8017c1ec,     //TAP_PlayPCM
//    0x80159588,     //process_eit
};


//.ini variables
dword code_start = CODE_START;
dword code_end = CODE_END;
dword hook_start = 0x80000000;
dword hook_end = 0x80002000;
dword time_start = 0;
dword time_end = 0;
int capture_duration = 60;
dword debug = 0;

//----------------------------------------------------------------------------
void oprintf(const void *fmt, ...);
void ffprintf(TYPE_File *Op, const void *fmt, ...);

//hooks
int hook1_enabled = 0;
int hook1_id = -2;
int apply_hook1();
int remove_hook1();
dword hook1(dword param1, dword param2);

typedef struct {
    dword caller_addr;
    dword callee_addr;
    dword count;
}type_profile;
type_profile *profile = 0;

#define MAX_ENTRIES 0x10000
int num_entries = 0;

void *profile_hash = 0;

void patch_profile_hook();
int capture_end = 0;

//----------------------------------------------------------------------------
//menu stuff
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {11, "xxx",    },

        {0, "Status", },
        {1, "Hook 80000000-80080000", },
        {2, "Hook 80080000-800d0000", },
        {3, "Hook 800d0000-80100000", },       //if this range is bigger, the picture breaks up ...
        {4, "Hook 80100000-80180000", },
        {5, "Hook 80180000-80257444", },
        {6, "Hook ini range", },
        {7, "Save Results", },
        {8, "Reset", },
        {9, "Remove", },

        {10, "Exit Tap", },
    },
};

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(TAP_TableAddress + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    //determine support for different firmwares
    ret = exTAP_Init();
    TAP_Print("\nexTAP_Init(), ret=%x", ret);
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", SysID, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    //initialise firebirdlib
    ret = InitTAPex();
    TAP_Print("\nInitTAPex(), ret=%x", ret);
    if (ret == 0) {
        sprintf(strLog, "%s(), InitTAPex(), ret=%x", __func__, ret);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    ini_read();
    patch_profile_hook();
    hook1_enabled = 0;
//    apply_hook1();

    main_menu.visible = 0;
    TSRCommanderInit(0, TRUE);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    DeleteMenu(&main_menu);

    remove_hook1();
//    TAP_Delay(10);                                             // show it for dwDelay /100 seconds

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);

    profile_remove();

    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    TAP_Exit();
    return 0;
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword state, subState;
    int tick;

    switch (event) {
        case EVT_IDLE:
            tick = TAP_GetTick();
            //
            //message window
            //
            if (message_showing) {
                if (tick > message_showing) {
                    TAP_Osd_Delete(message_rgn);
                    message_rgn = 0;
                    message_showing = 0;
                }
            }

            //stop after 1 minute
            if (profile) {
                if (tick > capture_end) {
                    profile_remove();
                }
            }

            TSRCommanderWork();
            break;
        case EVT_SVCSTATUS:
            break;

        case EVT_KEY:
            TAP_GetState(&state, &subState);
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
//            TAP_Print("\nstate = %s, subState = %s", get_str(state_str, state), get_str(substate_str, subState));
            if (main_menu.visible) {
                param1 = menu_handler(&main_menu, param1);
            }
            else {
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    param1 = key_handler(param1, param2);
                }
            }
            break;

        case EVT_HOOK:         //0xf100
            if (param1 == hook1_id) param1 = hook1(param1, param2);
            break;
        default:
            break;
    }

    return param1;
}


//----------------------------------------------------------------------------
int menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int num_events;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    profile_status();
				    return 0;
				case 1:
  				    profile_remove();
				    hook_start = 0x80000000;
				    hook_end   = 0x80080000;
				    setup();
				    return 0;
				case 2:
   				    profile_remove();
				    hook_start = 0x80080000;
				    hook_end   = 0x800d0000;
				    setup();
				    return 0;
				case 3:
   				    profile_remove();
				    hook_start = 0x800d0000;
				    hook_end   = 0x80100000;
				    setup();
				    return 0;
				case 4:
   				    profile_remove();
				    hook_start = 0x80100000;
				    hook_end   = 0x80180000;
				    setup();
				    return 0;
				case 5:
   				    profile_remove();
				    hook_start = 0x80180000;
				    hook_end   = code_end;
				    setup();
				    return 0;
				case 6:
   				    profile_remove();
				    ini_read();
				    setup();
				    return 0;
				case 7:
				    save_profile();
				    return 0;
				case 8:
				    reset();
				    return 0;
				case 9:
				    profile_remove();
    				return 0;

				case 10:
                    close();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}

int key_handler(dword param1, dword param2)
{
    int ret;

    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;
        case RKEY_Red:
            red();
            return 0;
            break;
        case RKEY_Green:
            green();
            return 0;
            break;
        case RKEY_Yellow:
            yellow();
            return 0;
            break;
        case RKEY_Blue:
            blue();
            return 0;
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//print multiline message to OSD
//  500=vertical position(0-576)
//  RGB(3,5,10)=background colour(dark blue)
//  200=hold time(2 seconds)
//  0=blocking
//
void oprintf(const void *fmt, ...)
{
    va_list argp;
    int newline_count;
    int i,y;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);

    newline_count = 0;
    for (i=0; i<strlen(strLog); i++) {
        if (strLog[i] == 0x0a) newline_count++;
    }
    y = 500 - (newline_count * 22);

    ShowMessage_y(strLog, y, RGB(3,5,10), 500, 0);
    va_end(argp);

    return;
}

void ffprintf(TYPE_File *Op, const void *fmt, ...)
{
    va_list argp;
    int len;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);

    len = strlen(strLog);
    TAP_Hdd_Fwrite(strLog, 1, len, Op);

    va_end(argp);

    return;
}
//----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//hook stuff
//
//hook the ???() function
//
//


type_hook hook1_patch[] = {
// sysid, firmware, enable, address,    length, original_data
    {456, 0x1359,   1,      0x80156A00, 8,      {0x27, 0xbd, 0xff, 0xe8,  0x8f, 0xaa, 0x00, 0x2c, }, },  //epg_cache
};

int apply_hook1(void)
{

    hook1_id = exTAP_ApplyHook(&hook1_patch[0]);
    if (hook1_id < 0) {
        hook1_enabled = 0;
        sprintf(strLog, "ERROR: apply_hook1(), firmware version %04x not supported", _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }

    hook1_enabled = 1;
    return 0;
}

int remove_hook1(void)
{
    int ret;

    if (hook1_id >=0) {
        ret = exTAP_RemoveHook(hook1_id);
        if (ret == 0) hook1_id = -1;
    }

    hook1_enabled = 0;
    return 0;
}

dword hook1(dword param1, dword param2)
{
    type_hook_registers *reg;
    int crc, service_id, transport_stream_id, original_network_id, section_number, version_number;
    dword ret;
    static int count = 0;
    byte *section;
    byte table_id = 0;

//TAP_Print("\nEVT_HOOK1, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
//    if (count < 4) {
//        TAP_Print("\nhook1(%x, %x)", reg->a0, reg->a1);
//        dump_reg(reg);

        //previous function, +0x28, a0=section buffer
//        TAP_Print("\nsection buffer");
//        dump(re->sp+0x90+0xe8, 0x10);
//        if (is_firmware(*(dword *)(reg->sp+0x90+0xe8))) dump(*(dword *)(reg->sp+0x90+0xe8), 0x40);
//        count ++;
//    }


    return param1;
}

//-----------------------------------------------------------------------------
char ini_filename[] = "profiler.ini";

int ini_read()
{
    int ret;

    ret = INIOpenFile(ini_filename);
    printf("\nINIOpenFile(), ret=%x", ret);

    code_start = INIGetHexDWord("code_start", CODE_START, 0x80000000, 0x81000000);
    printf("\nINIGetHexDWord(\"code_start\"), code_start=%x", code_start);

    code_end = INIGetHexDWord("code_end", CODE_END, 0x80000000, 0x81000000);
    printf("\nINIGetHexDWord(\"code_end\"), code_end=%x", code_end);

    hook_start = INIGetHexDWord("hook_start", 0x80000000, 0x80000000, 0x81000000);
    printf("\nINIGetHexDWord(\"hook_start\"), hook_start=%x", hook_start);

    hook_end = INIGetHexDWord("hook_end", 0x80000000, 0x80000000, 0x81000000);
    printf("\nINIGetHexDWord(\"hook_end\"), hook_end=%x", hook_end);

    capture_duration = INIGetInt("capture_duration", 60, 0, 24*60*60);
    printf("\nINIGetInt(\"capture_duration\"), capture_duration=%d", capture_duration);

    debug = INIGetHexDWord("debug", 0x00000000, 0x00000000, 0xffffffff);
    printf("\nINIGetHexDWord(\"debug\"), debug=%x", debug);

    INICloseFile();
    printf("\n");
}

//-----------------------------------------------------------------------------
static __inline__ int profiler();


int setup()
{
    int i;
    dword addr;
    dword data;
    int num_returns;
    int num_excluded;
    int num_included;
    int excluded;
    word mjd;
    byte hour, min, sec;


    num_included = sizeof(include) / 4;

    //count the number of functions, so that we can allocate enough memory
    num_returns = 0;
    for (addr=hook_start; addr<hook_end; addr+=4) {
        if (*(dword *)addr == 0x03e00008) {     //jr $ra
//            if (addr < 0x80010000) TAP_Print("\nfound subroutine at %08x=%08x, profiler()=%08x", addr, *(dword *)addr, profiler);
            num_returns++;
        }
    }
    //num_returns = ~7500
    //av. num calls per function = ~8 ?
    TAP_Print("\nMAX_ENTRIES=%x, found %d subroutines", MAX_ENTRIES, num_returns);

    //allocate the table
    if (profile) free(profile);
    profile = 0;
    profile = (type_profile *)malloc(MAX_ENTRIES * sizeof(type_profile));
    if (!profile) {
        TAP_Print("\nmalloc() failed");
        return 0;
    }
    for (i=0; i<=MAX_ENTRIES; i++) {
        profile[i].caller_addr = 0;
        profile[i].callee_addr = 0;
        profile[i].count = 0;
    }
    num_entries = 0;

    //initialise the hash
    if (profile_hash) close_hash(profile_hash);
    profile_hash = 0;
    profile_hash = open_hash(8, num_returns+num_included);       //match on the caller address, callee address

    //generate a timestamp
    TAP_GetTime(&mjd, &hour, &min, &sec);
    time_start = hour<<16 | min<<8 | sec;

    //modify the firmware
    TAP_Print("\nadding %d hooks, %x-%x", num_returns, hook_start, hook_end);
    oprintf("adding %d hooks, %x-%x", num_returns, hook_start, hook_end);
    data = 0x08000000 | (((dword)profiler >> 2) & 0x03ffffff);  //j profiler
    for (addr=hook_start; addr<hook_end; addr+=4) {
        if (*(dword *)addr == 0x03e00008) {     //jr $ra
//exclusions
//interrupts?
//some functions may be called in a taps $gp context ...?
            num_excluded = sizeof(exclude) / 4;
            excluded = 0;
            for (i=0; i<num_excluded; i++) {
                if (addr == exclude[i]) {
                    excluded = 1;
                    break;
                }
            }
            if (excluded) {
                //skip any explictly excluded
            }
            else {      //(not_in_exclusion_list) {
//                TAP_Print("\n*%08x = %08x", addr, data);
                *(dword *)addr = data;          //j profiler
            }
        }
    }

    for (i=0; i<num_included; i++) {
        addr = include[i];
        if (*(dword *)addr == 0x03e00008) {     //jr $ra
            TAP_Print("\n*%08x = %08x", addr, data);
            *(dword *)addr = data;          //j profiler
        }
        else {
            TAP_Print("\noops, *%08x = %08x", addr, *(dword *)addr);
//            *(dword *)addr = data;          //j profiler
        }
    }

    //run for one minute
    capture_end = TAP_GetTick() + 100 * capture_duration;

    return 0;
}

int stop()
{
    int i;
    dword addr;
    dword data;
    int num_excluded;
    int num_included;
    int excluded;
    word mjd;
    byte hour, min, sec;

    //generate a timestamp
    TAP_GetTime(&mjd, &hour, &min, &sec);
    time_end = hour<<16 | min<<8 | sec;

    data = 0x08000000 | (((dword)profiler >> 2) & 0x03ffffff);  //j profiler

    //to restore the firmware ...
    num_excluded = sizeof(exclude) / 4;
    for (addr=hook_start; addr<hook_end; addr+=4) {
        if (*(dword *)addr == data) {           //j profiler
            excluded = 0;
            for (i=0; i<num_excluded; i++) {
                if (addr == exclude[i]) {
                    excluded = 1;
                    break;
                }
            }
            if (excluded) {
                //skip any explictly excluded
            }
            else {  //if (not_in_exclusion_list) {
                *(dword *)addr = 0x03e00008;    //jr $ra
            }
        }
    }

    num_included = sizeof(include) / 4;
    for (i=0; i<num_included; i++) {
        addr = include[i];
        if (*(dword *)addr == data) {           //j profiler
            *(dword *)addr = 0x03e00008;    //jr $ra
        }
        else {
            TAP_Print("\noops, *%08x = %08x", addr, *(dword *)addr);
//            *(dword *)addr = 0x03e00008;    //jr $ra
        }
    }
}

int profile_remove()
{

    if (profile == 0) return 0;

    TAP_Print("\nremoving all hooks, saving profile, num_entries=%d", num_entries);
    oprintf("removing all hooks, saving profile, num_entries=%d", num_entries);

    stop();
    if (num_entries) save_profile();

    if (profile_hash) close_hash(profile_hash);
    profile_hash = 0;

    if (profile) free(profile);
    profile = 0;

    num_entries = 0;

    return 0;
}

int reset()
{
    int i;

    //reset all the counters
    TAP_Print("\nresetting all counts");
    oprintf("resetting all counts");
    for (i=0; i<num_entries; i++) {
        profile[i].count = 0;
    }

    return 0;
}

int profile_status()
{
    int i;

    TAP_Print("\nnum_entries = %d", num_entries);
    oprintf("num_entries = %d", num_entries);
    for (i=0; ((i<num_entries) && (i<20)); i++) {
        TAP_Print("\n%08x %08x %d", profile[i].caller_addr, profile[i].callee_addr, profile[i].count);
    }

    return 0;
}

int save_profile()
{
    word mjd, year;
    byte month, day, weekDay;
    byte hour, min, sec;
    int ret;
    int i;
    char filename[0x40];
    TYPE_File *Op;

    //create the filename based on date,time
    TAP_GetTime(&mjd, &hour, &min, &sec);
    TAP_ExtractMjd(mjd, &year, &month, &day, &weekDay);
    sprintf(filename, "profile_%04d%02d%02d%02d%02d%02d.dat", year, month, day, hour, min, sec);
    TAP_Print("\nsaving %s, %x-%x, num_entries=%d", filename, hook_start, hook_end, num_entries);
    oprintf("saving %s, %x-%x, num_entries=%d", filename, hook_start, hook_end, num_entries);

    ret = TAP_Hdd_ChangeDir("/");
    ret = TAP_Hdd_ChangeDir("ProgramFiles");
    if (TAP_Hdd_Exist(filename)) {
        TAP_Print("\n%s exists, delete it", filename);
        TAP_Hdd_Delete(filename);
    }
    TAP_Hdd_Create(filename, ATTR_NORMAL);
    Op = TAP_Hdd_Fopen(filename);
    ret = TAP_Hdd_Fseek(Op, 0, SEEK_SET);

    ffprintf(Op, "profiler\n");
    ffprintf(Op, "%x - %x\n", hook_start, hook_end);
    ffprintf(Op, "%02d:%02d:%02d - %02d:%02d:%02d\n", (time_start>>16)&0xff, (time_start>>8)&0xff, (time_start>>0)&0xff, (time_end>>16)&0xff, (time_end>>8)&0xff, (time_end>>0)&0xff);
    ffprintf(Op, "num_entries=%d\n", num_entries);
    ffprintf(Op, "caller_addr callee_addr    count\n");
    for (i=0; i<num_entries; i++) {
        ffprintf(Op, " %08x    %08x %8d\n", profile[i].caller_addr, profile[i].callee_addr, profile[i].count);
    }
    ffprintf(Op, "-------- end of file --------\n");

    TAP_Hdd_Fclose(Op);
    return 0;
}

//-----------------------------------------------------------------------------
extern dword *profile_hook_main;
extern void *hook_entry_profile;
int profiler1();

//
//no need to save all the regs ...
//
static __inline__ int hook_prefix()
{
    //save some registers
    __asm__ __volatile__ (
    ".set noat\n"
    "addiu   $29,$29,0xff70\n"              //addiu   $sp,$sp,-0x90
//    "sw      $0, 0x0010($29)\n"             //sw      $0, 0x0010($sp)         save 0
//    "sw      $1, 0x0014($29)\n"             //sw      $at,0x0014($sp)         save at
    "sw      $2, 0x0018($29)\n"             //sw      $v0,0x0018($sp)         save v0
    "sw      $3, 0x001c($29)\n"             //sw      $v1,0x001c($sp)         save v1
    "sw      $4, 0x0020($29)\n"             //sw      $a0,0x0020($sp)         save a0
    "sw      $5, 0x0024($29)\n"             //sw      $a1,0x0024($sp)         save a1
    "sw      $6, 0x0028($29)\n"             //sw      $a2,0x0028($sp)         save a2
    "sw      $7, 0x002c($29)\n"             //sw      $a3,0x002c($sp)         save a3

//    "sw      $8, 0x0030($29)\n"             //sw      $t0,0x0030($sp)         save t0
//    "sw      $9, 0x0034($29)\n"             //sw      $t1,0x0034($sp)         save t1
//    "sw      $10,0x0038($29)\n"             //sw      $t2,0x0038($sp)         save t2
//    "sw      $11,0x003c($29)\n"             //sw      $t3,0x003c($sp)         save t3
//    "sw      $12,0x0040($29)\n"             //sw      $t4,0x0040($sp)         save t4
//    "sw      $13,0x0044($29)\n"             //sw      $t5,0x0044($sp)         save t5
//    "sw      $14,0x0048($29)\n"             //sw      $t6,0x0048($sp)         save t6
//    "sw      $15,0x004c($29)\n"             //sw      $t7,0x004c($sp)         save t7
//
    "sw      $16,0x0050($29)\n"             //sw      $s0,0x0050($sp)         save s0
    "sw      $17,0x0054($29)\n"             //sw      $s1,0x0054($sp)         save s1
    "sw      $18,0x0058($29)\n"             //sw      $s2,0x0058($sp)         save s2
    "sw      $19,0x005c($29)\n"             //sw      $s3,0x005c($sp)         save s3
//    "sw      $20,0x0060($29)\n"             //sw      $s4,0x0060($sp)         save s4
//    "sw      $21,0x0064($29)\n"             //sw      $s5,0x0064($sp)         save s5
//    "sw      $22,0x0068($29)\n"             //sw      $s6,0x0068($sp)         save s6
//    "sw      $23,0x006c($29)\n"             //sw      $s7,0x006c($sp)         save s7
//
//    "sw      $24,0x0070($29)\n"             //sw      $t8,0x0070($sp)         save t8
//    "sw      $25,0x0074($29)\n"             //sw      $t9,0x0074($sp)         save t9
//    "sw      $26,0x0078($29)\n"             //sw      $k0,0x0078($sp)         save k0
//    "sw      $27,0x007c($29)\n"             //sw      $k1,0x007c($sp)         save k1
    "sw      $28,0x0080($29)\n"             //sw      $gp,0x0080($sp)         save gp
//    "sw      $29,0x0084($29)\n"             //sw      $sp,0x0084($sp)         save sp
//    "sw      $30,0x0088($29)\n"             //sw      $s8,0x0088($sp)         save s8
    "sw      $31,0x008c($29)\n"             //sw      $ra,0x008c($sp)         save ra
    :
    :
    );
}

static __inline__ hook_wrapper()
{
   __asm__ __volatile__ (
    ".set noat\n"
    "hook_main:\n"

    "profile_hook_main:         \n"     //    a label so we can find this location to patch it
    "li    $18, 0x8017a0d0      \n"     //    li      $s2, FW_TAP_TABLE_1359         !!! firmware specific,  patch this location later !!!
    "li    $17, 0x8017a2d0      \n"     //    li      $s1, FW_TAP_IDENT_1359         !!! firmware specific,  patch this location later !!!
    "li    $5,  0xeeeeeeee      \n"     //    li      $a1, tap_id                    !!! runtime specific,  patch this location later !!!
    "li    $19, 0x88888888      \n"     //    li      $s3, profiler1                 !!! runtime specific,  patch this location later !!!

    "lw    $16, 0($17)          \n"     //    lw      $s0, 0($s1)            //save initial tap_ident

    "sll   $5,  5               \n"     //    sll     $a1, 5
    "sw    $5,  0($17)          \n"     //    sw      $a1, 0($s1)            //set current tap_ident

    "lw    $5,  4($18)          \n"     //    lw      $a1, 4($s2)            //tap gp
    "move  $28, $5              \n"     //    move    $gp, $a1               //set gp

    "addiu $4,  $29, 0x10       \n"     //    addiu   $a0, $sp, 0x10         //param = pointer to saved registers
    "jalr  $19                  \n"     //    jalr    $s3                    //profiler1()
//nop

    "sw    $16, 0($17)          \n"     //    sw      $s0, 0($s1)            //restore initial tap_ident
    :
    :
    );
}


//k0=
//k1 = exception pc

static __inline__ int hook_postfix()
{
    //restore the registers
    __asm__ __volatile__ (
    ".set noat\n"
    "lw      $31,0x008c($29)\n"             //lw      $ra,0x008c($sp)         restore ra
//    "lw      $30,0x0088($29)\n"             //lw      $s8,0x0088($sp)         restore s8
//  "lw      $29,0x0084($29)\n"             //lw      $sp,0x0084($sp)         restore sp
    "lw      $28,0x0080($29)\n"             //lw      $gp,0x0080($sp)         restore gp
//    "lw      $27,0x007c($29)\n"             //lw      $k1,0x007c($sp)         restore k1
//    "lw      $26,0x0078($29)\n"             //lw      $k0,0x0078($sp)         restore k0
//    "lw      $25,0x0074($29)\n"             //lw      $t9,0x0074($sp)         restore t9
//    "lw      $24,0x0070($29)\n"             //lw      $t8,0x0070($sp)         restore t8

//    "lw      $23,0x006c($29)\n"             //lw      $s7,0x006c($sp)         restore s7
//    "lw      $22,0x0068($29)\n"             //lw      $s6,0x0068($sp)         restore s6
//    "lw      $21,0x0064($29)\n"             //lw      $s5,0x0064($sp)         restore s5
//    "lw      $20,0x0060($29)\n"             //lw      $s4,0x0060($sp)         restore s4
    "lw      $19,0x005c($29)\n"             //lw      $s3,0x005c($sp)         restore s3
    "lw      $18,0x0058($29)\n"             //lw      $s2,0x0058($sp)         restore s2
    "lw      $17,0x0054($29)\n"             //lw      $s1,0x0054($sp)         restore s1
    "lw      $16,0x0050($29)\n"             //lw      $s0,0x0050($sp)         restore s0

//    "lw      $15,0x004c($29)\n"             //lw      $t7,0x004c($sp)         restore t7
//    "lw      $14,0x0048($29)\n"             //lw      $t6,0x0048($sp)         restore t6
//    "lw      $13,0x0044($29)\n"             //lw      $t5,0x0044($sp)         restore t5
//    "lw      $12,0x0040($29)\n"             //lw      $t4,0x0040($sp)         restore t4
//    "lw      $11,0x003c($29)\n"             //lw      $t3,0x003c($sp)         restore t3
//    "lw      $10,0x0038($29)\n"             //lw      $t2,0x0038($sp)         restore t2
//    "lw      $9, 0x0034($29)\n"             //lw      $t1,0x0034($sp)         restore t1
//    "lw      $8, 0x0030($29)\n"             //lw      $t0,0x0030($sp)         restore t0

    "lw      $7, 0x002c($29)\n"             //lw      $a3,0x002c($sp)         restore a3
    "lw      $6, 0x0028($29)\n"             //lw      $a2,0x0028($sp)         restore a2
    "lw      $5, 0x0024($29)\n"             //lw      $a1,0x0024($sp)         restore a1
    "lw      $4, 0x0020($29)\n"             //lw      $a0,0x0020($sp)         restore a0
    "lw      $3, 0x001c($29)\n"             //lw      $v1,0x001c($sp)         restore v1
    "lw      $2, 0x0018($29)\n"             //lw      $v0,0x0018($sp)         restore v0
//    "lw      $1, 0x0014($29)\n"             //lw      $at,0x0014($sp)         restore at
//  "lw      $0, 0x0010($29)\n"             //lw      $0, 0x0010($sp)         restore 0
    "addiu   $29,$29,0x0090\n"              //addiu   $sp,$sp,0x90
    "hook_postfix2:\n"
    :
    :
    );
}

static __inline__ int profile_hook_exit()
{
    __asm__ __volatile__ (
    "j $31                  \n"     //    j     $ra
    "nop                    \n"     //    nop
    );
}

void patch_profile_hook()
{
    //set the address of tap_ident, tap_table in hook_main()
    *(word *)((dword)&profile_hook_main+0x02) = ((dword)fw_tap_table >> 16) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x06) = ((dword)fw_tap_table >> 00) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x0a) = ((dword)fw_tap_ident >> 16) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x0e) = ((dword)fw_tap_ident >> 00) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x12) = (*(dword *)fw_tap_ident >> 16) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x16) = (*(dword *)fw_tap_ident >> 00) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x1a) = ((dword)profiler1 >> 16) & 0xffff;
    *(word *)((dword)&profile_hook_main+0x1e) = ((dword)profiler1 >> 00) & 0xffff;
}


static __inline__ int profiler()
{
    //cannot use anything that relies on $gp here ...
    //i.e. constants only, no variables

    __asm__ __volatile__ ("hook_entry_profile:\n");
    hook_prefix();      //save the original resgiters
    hook_wrapper();     //fixes $gp, calls user code
    hook_postfix();     //restores the original registers
    profile_hook_exit();        //rts
}

//-----------------------------------------------------------------------------
//
// need to preserve $v0
//

    //save registers
    //switch $gp context
    //  $ra -> caller address
    //  *($ra-8) -> callee address
    //      match caller, callee address in hash
    //      if not found, add it
    //      increment count
    //switch $gp context
    //restore registers
    //do jr $r


int profiler1(dword param)
{
    type_hook_registers *reg;
    dword caller_addr, callee_addr;

//TAP_Print("\nprofiler1, param=%x", param);

    reg = (type_hook_registers *)param;
//    dump_reg(reg);
//    dump32(reg->sp-0x80, 0x200);
//    TAP_Print("\n");
//    dump32(reg->sp+0x90, 0x40);

//what if
//      000000ss_sss00000_dddddhhh_hh001001    jalr $s0
    caller_addr = reg->ra-0x8;
    if (((*(dword *)caller_addr) & 0xfc00003f) == 0x00000009) {
        callee_addr = 0;
//        dump_reg(reg);
//        TAP_Print("\n");
//        dump32(reg->sp-0x80, 0x200);
    }
    else {
        callee_addr = 0x80000000 | (((*(dword *)caller_addr) & 0x03ffffff) << 2);
    }

//    if (callee_addr)
    profiler2(caller_addr, callee_addr);
    return 0;
}

//match caller, callee address in hash
//if not found, add it
//increment count
int profiler2(dword caller_addr, dword callee_addr)
{
    type_profile profile1;
    type_hash_element *hashptr;
    int index;

if (debug & 1) TAP_Print("\ncaller_addr=%x, callee_addr=%x", caller_addr, callee_addr);
    profile1.caller_addr = caller_addr;
    profile1.callee_addr = callee_addr;
    profile1.count = 0;

    //match the table
    hashptr = findhash(profile_hash, (byte *)&profile1);

    //not in there
    if (!hashptr) {
        index = num_entries;
        profile[index] = profile1;
        num_entries++;
        if (num_entries == MAX_ENTRIES) {
            //ignore it
TAP_Print("\nfull; num_entries=%x", num_entries);
            num_entries --;
            return num_entries;
        }
if (debug & 1) TAP_Print(", add,   index=%x", index);
        addhash(profile_hash, (byte *)&profile[index], index);
    }
    else {
        //found it
        index = hashptr->data;
if (debug & 1) TAP_Print(", found, index=%x", index);
    }

    //update it
    profile[index].count++;
if (debug & 1) TAP_Print(",count=%x, num_entries=%x", profile[index].count, num_entries);

    return num_entries;
}


//-----------------------------------------------------------------------------
//
//testing
//
int red() {
    profile_remove();
    return 0;
}

int green() {
    return 0;
}

int yellow() {
    return 0;
}

int blue() {
    TAP_Print("\nTAP_PlayPCM");
    TAP_PlayPCM(0, 0, 1000, 0);
    return 0;
}

